# Spatial restoration / forward script
#   * Grids must be in the same reference coordinate system
#   * Use TIF files with world files, make sure data type is 32-bit or lower
#   * 1. Grid which is to deform
#   * 2. Grid with strikes (direction of displacement)
#   * 3. Grid with amount of displacement (m) / vector size
#   * Grid 2 and 3 should have the same extent and resolution for best results
#   * If you want to have a video, make sure ffmpeg is installed and accessible
#   * Image deformations are done with the PIL library
# -- Marc Schori, 2018
#
# How to install PIL library:
#       python.exe -m pip install Pillow
#       python.exe -m pip install libtiff
#___________________________________________________________

import numpy
import copy
import datetime
import time
import subprocess
import ntpath
import math
import sys
from ConfigParser import SafeConfigParser
import glob
import os


from PIL import Image
from PIL import ImageFont
from PIL import ImageDraw
from PIL import TiffImagePlugin


# ----------------------------------
# Class loading georeferenced image (TIF) <= 32-Bit
# ----------------------------------
class spatialImage:
    
    name = ''

    # Information from Worldfile
    x_coord = float(0)
    y_coord = float(0)
    x_cellsize = float(0)
    y_cellsize = float(0)
    x_skew = float(0)
    y_skew = float(0)
    orig_tfw = ""

    def px2m (self, px):
        return px * self.x_cellsize

    def m2px (self, m):
        return int(round(float(m) / self.x_cellsize,0))
        

    # Will contain PIL Image object
    # http://pillow.readthedocs.io/en/3.4.x/reference/Image.html
    im = None
    
    def __init__(self, name):
        self.name = name
        self.loadImage()
        self.loadWorldfile()

    def save(self):
        print "Saving " + self.name
        wf = open(self.name + '.tfw', "w")
        wf.write(self.orig_tfw)
        wf.close()
        #TiffImagePlugin.WRITE_LIBTIFF = True

        # You might have to play a bit. Not every compression is available, depending
        # on libtiff library (make sure you install it with pip for example).
        #self.im.save(self.name + '.tif', compression='tiff_jpeg')
        #self.im.save(self.name + '.tif', compression='tiff_adobe_deflate')
        #self.im.save(self.name + '.tif', compression='tiff_ccitt')
        self.im.save(self.name + '.tif', compression='tiff_deflate')
        #self.im.save(self.name + '.tif')

    # +--------------------------------+
    # Draw scale, frame, grid and title text
    # +--------------------------------+
    def imprint_scale(self, text = ""):
        frame_width = 15
        border_spacing = 10 + frame_width
        line_spacing = float(font_size) * 0.4
        scale_px = myImage.m2px(scale_size * 1000)


        grid_color = (100,100,100)
        #white =  (255,255,255)
        #black = (0,0,0)

        x_max=self.im.size[0]
        y_max=self.im.size[1]
        x1 = border_spacing
        x2 = x1 + scale_px / 2
        x3 = x2 + scale_px / 2
        y2 = y_max - border_spacing
        y1 = y2 - (float(font_size) * 0.6)         # bar height as ratio of text height
        y_text = y1 - font_size - line_spacing     # text distance from bar

        draw = ImageDraw.Draw(self.im)

        # Make sure font is in current directory
        font = ImageFont.truetype("arial.ttf", font_size)

     
                       
        # --- grid
        if grid_spacing > 0:
            dist = self.m2px(grid_spacing * 1000)
            for x in xrange(frame_width+dist, x_max, dist):
                draw.line([x, frame_width, x, y_max - frame_width], fill=grid_color, width=1)

            for y in xrange(frame_width+dist, y_max, dist):
                draw.line([frame_width, y, x_max - frame_width, y], fill=grid_color, width=1)
            
        # --- frame
        draw.rectangle([1,1,frame_width,y_max], (255,255,255), (255,255,255))
        draw.rectangle([1,1,x_max, frame_width], (255,255,255), (255,255,255))
        draw.rectangle([x_max - frame_width,1,x_max,y_max], (255,255,255), (255,255,255))
        draw.rectangle([1,y_max-frame_width,x_max,y_max], (255,255,255), (255,255,255))
        draw.rectangle([frame_width,frame_width,x_max-frame_width,y_max-frame_width], None,(0,0,0))

        # --- Draw scale
        draw.rectangle([x1,y1,x2,y2], (0,0,0), (0,0,0))
        draw.rectangle([x2,y1,x3,y2], (255,255,255), (0,0,0))
        draw.text((x1, y_text), "0", 0, font=font)
        draw.text((x3, y_text), str(scale_size) + " km", 0, font=font)

        # --- Title
        draw.text((border_spacing, border_spacing),text,0,font=font)

    # -------------------
    # worldfile syntax
    # 0. resolution x (x-scale)
    # 1. rotation about y-axis (y-skew)
    # 2. rotation about x-axis (x-skew)
    # 3. resolution y (y-scale, negative value)
    # 4. x-coordinate of the center of the upper left pixel
    # 5. y-coordinate of the center of the upper left pixel
    
    def loadWorldfile(self):
        wf = open(self.name + '.tfw', "r")
        self.orig_tfw = wf.read()
        wf.close()
        lines = self.orig_tfw.split("\n")
        lines = [x.strip() for x in lines]
        
        self.x_coord = float(lines[4])
        self.y_coord = float(lines[5])
        self.x_cellsize = float(lines[0])
        self.y_cellsize = float(lines[3])
        self.y_skew = float(lines[1])
        self.x_skew = float(lines[2])
        
        return lines
    
    def loadImage(self):
        im = Image.open(self.name + '.tif', mode='r')
        #w, h = im.size
        #print w
        #print h
        #im = im.crop((0,0,w,h+1000))
        #im.show()
        self.im = im
        

        print "  Grid name: " + self.name
        print "  Extent:    " + str(im.size[0]) + "x" + str(im.size[1])
        print "  Min - Max: " + str(im.getextrema()[0]) + " - " + str(im.getextrema()[1])

        i = 0
        for abb in im.getbands():
            i= i + 1
            print "  Band " + str(i) + ":    " + modes(abb)

        print "Mode:" + im.mode
        #print im.info
        #print "Format:    " + im.format

        #ba = im.getbands()
        #print "Bands:" + str(ba)
        #co = im.getcolors()
        #print "Colors:" + str(co)
        #hi = im.histogram()
        #print "Histogram:" + str(hi)

        print
        return im

def isNaN(num):
    return num != num

# --------------------------
# conditional function for image range
# --------------------------
def inGrid(x,y, im):
    if (x>=1 and y>=1 and x<im.size[0] and y<im.size[1]):
        return True
    else:
        return False

# ----------------------------------------------
# Invert strikes
# ----------------------------------------------
def strike_inv(strike):
    strike = strike + 180
    if strike > 359:
        strike = strike - 360
    return strike

# --------------------------
# Function returning new position after adding vector
# using geographic circle with N = 0 degree
# parameter max can be used for incremental displacement
# -------------------------
def trans(x_coord, y_coord, strike, length, max=0, model_type=0):

    # Do not alter pixel if length or strike are not defined
    if length <= 0 or strike < 0:
        return (x_coord,y_coord)

    # Constant displacement speed
    if model_type == 0:
        if length > max:
            length = max  # full distance
            
    # Pervasive displacement with varying speeds
    else:
        perc = float(max) / float(max_displacement)
        length = perc * length  # percentage

    # Invert strikes if specified
    if invert_strikes:
        strike = strike_inv(strike)
        
    alpha = strike * numpy.pi / 180
    x_new = x_coord + (numpy.sin(alpha) * length)
    y_new = y_coord + (numpy.cos(alpha) * length)
    return (x_new,y_new)


def progress_bar(x,y, xmax, ymax):
    global previous
    z = float(x*y)
    n = float(xmax*ymax)
    nr = float(z / n * 100)
    if int(nr) % 5 == 0 and int(nr)  > previous:
        previous = int(nr)
        sys.stdout.write('.')

# ----------------------------------------------
# Translate pixels of MYIMAGE acc. to step (max)
# ----------------------------------------------
# authors note: it seems that PIL has troubles to read 32-bit float grids.
#   The lowermost part of the grid will be cut away. Hence, add security space
#   to your image if you need to include everything

def deform_image(max=0, model_type=0, sequence_number=0):

    global previous
    previous = 0

    # Information printing and statistic numbers
    startdate = datetime.datetime.now().time()
    starttime = time.time()
    print str(startdate) + " Calculate displacement at " + str(max) + " m ..."
   
    # In case the mode is F, L, P (all 1-band images), we probably have a grid
    # normal colour definitions do not apply there, hence we need to remove some functions
    global bgcolor
    global imprint_scale
    if myImage.im.mode[0] == 'F' or myImage.im.mode[0] == 'L' or myImage.im.mode[0] == 'P':
        bgcolor = 0
        imprint_scale = False
    
    # Create an empty image with the same extent as MYIMAGE
    # and copy world file information. Also copy palette if 8-bit
    myOutput = Image.new(myImage.im.mode,myImage.im.size, bgcolor)
    if (myImage.im.mode[0] == 'L' or myImage.im.mode[0] == 'P'):
        if (myImage.im.getpalette() != None):
            myOutput.putpalette(myImage.im.getpalette())

    # This grid contains the displaced length grid in the extent of myImage
    # -> in case of pixel overlap, this grid helps to decide what pixel value will be chosen
    myLengthRef = Image.new(myLength.im.mode,myImage.im.size, -1)

    # The bgcolor might change depending on how many bands myImage has.
    # Therefore, we read the first px in order to have an accurate definition
    bgcolor_enhanced = myOutput.getpixel((1,1))

    # The loop through every pixel in myImage: bottom-up and left-right
    for x in range (0, myImage.im.size[0]):
        for y in range (0, myImage.im.size[1]):

            progress_bar(x,y,myImage.im.size[0],myImage.im.size[1] )
            
            # calculate coordinates corresponding to pixel position in MYIMAGE
            # and read the color value
            posx = myImage.x_coord + x * myImage.x_cellsize
            posy = myImage.y_coord + y * myImage.y_cellsize
            color = myImage.im.getpixel((x,y)) # store pixel value

            # find the corresponding pixel position in the transformation grids from myImage coordinates.
            x_trans_strike = round((posx - myStrike.x_coord) / myStrike.x_cellsize,0)
            y_trans_strike = round((posy - myStrike.y_coord) / myStrike.y_cellsize,0)
            x_trans_length = round((posx - myLength.x_coord) / myLength.x_cellsize,0)
            y_trans_length = round((posy - myLength.y_coord) / myLength.y_cellsize,0)

            # ... and get values from strike and length grid (Length in [m] and strike in [degree])
            gStrike = -1
            gLength = 0
            if inGrid(x_trans_strike, y_trans_strike, myStrike.im):
                gStrike = myStrike.im.getpixel((x_trans_strike,y_trans_strike))
            if inGrid(x_trans_length, y_trans_length, myLength.im):
                gLength = myLength.im.getpixel((x_trans_length,y_trans_length))

            # Calculate coordinate (in m) after displacement, based on myImage positions.
            # (Trans function uses trigonometry to calculate position)
            newpos = trans(posx,posy,gStrike,gLength, max, model_type)
            xc2 = newpos[0]
            yc2 = newpos[1]

            if not isNaN(xc2) and not isNaN(yc2):
                
                # Calculate new pixel position in MYIMAGE from new coordinates (xc2, yc2)
                x2 = int(round((xc2 - myImage.x_coord) / myImage.x_cellsize,0))
                y2 = int(round((yc2 - myImage.y_coord) / myImage.y_cellsize,0))
            
                # Write pixel to new position and store length as well in a separate grid
                # ... but respect some rules:
                if inGrid(x2,y2,myOutput):

                    # A value was written before on the new position -> If overlap marking
                    # is enabled we mark this cell with a special colour.
                    if mark_overlap and (myOutput.getpixel((x2,y2)) != bgcolor_enhanced):
                        myOutput.putpixel((x2,y2), overlapcolor)
                        myLengthRef.putpixel((x2,y2), gLength)
                        
                    # A value was written before on the new position -> Only store new value
                    # if the current displacement is actually larger than the one already written
                    elif myOutput.getpixel((x2,y2)) != bgcolor_enhanced:
                        if gLength > myLengthRef.getpixel((x2,y2)):
                            myOutput.putpixel((x2,y2), color)
                            myLengthRef.putpixel((x2,y2), gLength)                    
                            
                    # No value on the new position -> store the current color and length
                    elif myOutput.getpixel((x2,y2)) == bgcolor_enhanced:
                        myOutput.putpixel((x2,y2), color)
                        myLengthRef.putpixel((x2,y2), gLength)
                        
    # Finally create a proper spatial object with new Image and save to file
    myNewImage = copy.copy(myImage)
    myNewImage.im = myOutput
    myNewImage.name = "results/" + ntpath.basename(myImage.name) + "_" + str(sequence_number).zfill(3)

    # Add text and frame incl. grid to the image
    if age_show:
        percentage = float(max)/float(max_displacement)
        #age_passed = round(float(age_span) * float(percentage),1)
        if not invert_strikes:
            # calculating age according to specified parameters (in case of a restoration grid)
            age_passed = age_span - round(float(age_span) * float(percentage),1)
            age_now = float(age_start) - float(age_passed)           
        else:
            # calculating age for forward grid
            age_passed = round(float(age_span) * float(percentage),1)
            age_now = float(age_start) - float(age_passed)
            
        caption_addon = ", " + str(age_now) + " " + str(age_unit)
    else:
        caption_addon = ''
    if model_type == 0:
        myText = "Displacement " + str(round(float(max)/1000.0,1)) + " km" + caption_addon
    elif model_type == 1:
        myText = "Displacement " + str(int(round(float(max)/float(max_displacement)*100,0))) + "%" + caption_addon

    # Only imprint scale if this is specified
    if imprint_scale:
        myNewImage.imprint_scale(myText)
    myNewImage.save()

    # Print time statistics
    endtime = time.time()
    diff = endtime - starttime
    steps = max / stepsize
    tleft = round(float(diff) * float(steps),1)
    sdiff = str(datetime.timedelta(seconds=round(diff,0)))
    stleft = str(datetime.timedelta(seconds=round(tleft,0)))
    print "Time used " + sdiff + ", time left " + stleft

    print
    return myNewImage

# +---------------------------------
# using ffmpeg, make sure that it is
# referenced in PATH environment variables
# Using mpeg4 codec which seems to be the fastest
# +---------------------------------
def createVideo():
    print "create video ..."
    ret = subprocess.call([
        'ffmpeg',
        '-framerate', str(framerate),
        '-f', 'image2',
        '-i', 'results/' + ntpath.basename(myImage.name) + '_%03d.tif',
        '-c:v', 'mpeg4',
        '-qscale:v', str(video_quality),
        'results/' + ntpath.basename(myImage.name) + '.mp4',
        '-y'
    ])

    if ret:
        print "Failure occurred!"
        return False
    else:
        print "Video saved in " + 'results/' + ntpath.basename(myImage.name) + '.mp4'
        return True

# ---------------------
# description of TIF modes
# ---------------------
def modes(abb):
    return {
        'I': "32-bit signed integer pixels",
        'F': "32-bit floating point pixels",
        '1': "1-bit pixels, black and white, stored with one pixel per byte",
        'L': "8-bit pixels, black and white",
        'P': "8-bit pixels, mapped to any other mode using a color palette",
        'RGB': "3x8-bit pixels, true color",
        'RGBA': "4x8-bit pixels, true color with transparency mask",
        'CMYK': "4x8-bit pixels, color separation",
        'YCbCr': "3x8-bit pixels, color video format",
        'LAB': "3x8-bit pixels, the L*a*b color space",
        'HSV': "3x8-bit pixels, Hue, Saturation, Value color space"
    }.get(abb, abb)


# -----------------------------------
# functions to store and read values
# -----------------------------------
def read_value(key, section = 'main'):
    config = SafeConfigParser()
    config.read('config.ini')

    try:
        s = config.get(section, key)
    except:
        s = ''
    
    return s

def store_value(key, value, section = 'main'):
    config = SafeConfigParser()
    config.read('config.ini')

    try:
        config.add_section(section)
    except:
        pass

    try:
        config.set(section, key, str(value))
    except:
        print(sys.exc_info()[0])
        
    with open('config.ini', 'w') as f:
        config.write(f)


# -----------------------------------
# Quick function for user interaction
# -----------------------------------
def ask(text, key):
    value = read_value(key)
    text = text.replace("%value%", value)

    x = raw_input(text)
    if x == '':
        x = value
    else:
        store_value(key,x)
        
    if x.isdigit():
        return int(x)

    return x

# ===================
# MAIN
# ====================

previous = 0
stepsize = 500              # stepsize in (m), 0 = no steps
bgcolor = (255,255,255)     # color for empty pixels (extension)
phcolor = (255,255,255)     # color for neutral pixels
overlapcolor = (255,200,200)# color for overlaps during unfolding (compression)
mark_overlap = False        # mark overlaps
model_type = 1              # 0 = meter steps, 1 = percentage steps
framerate = 20              # define video framerate
video_quality = 5           # video quality (1 = best, 31 = worst)
scale_size = 20            # scale size in km, use 10 as default
font_size = 25              # font size in px
imprint_scale = True        # Decides if any text / grid / scale should be added to the image
grid_spacing = 0            # grid spacing in km (0 = No grid), use 10 as default
sequence =  1               # 0: forward (0 to 100%), 1: backward (100% to 0)
invert_strikes = True       # Defines if given strikes are inverted (+180 degree)
age_span = 11               # Define timespan of model
age_start = 15              # Define start of deformation 
age_unit = 'Ma'             # Define age unit (Ma best)
age_show = False             # show age in text


print "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"
print "@          MAP RESTORATION          @"
print "@                and                @"
print "@         FORWARD MODELLING         @"
print "@               V1.03               @"
print "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
print "Hint: Make sure maps are stored as a TIF file using CH1903 / LV03 projection."
print "      An eponymous worldfile (tfw) needs to be annexed.\n"

#print os.path.dirname(__file__)
#print sys.path[0]
#os.chdir(os.path.dirname(sys.path[0])
os.chdir(os.path.dirname(__file__))

# --- Ask for user input on map choice
maps = glob.glob("maps/*.tif")
i = 0
# Make sire that maps are in CH1903 / LV03, TIF format and have a tfw world file.
print "Maps found in folder:"
for map in maps:
    print "  (" + str(i) + ") " + ntpath.basename(map)
    i = i + 1

x = ask("\nWhich map do you want to use [%value%] ? ", 'map')
input_map = maps[x]
input_map = os.path.splitext(input_map)[0]

# --- Ask for forward or restoration modelling
#model = ask('\nDo you want to run a (1) palinspastic restoration or (2) a forward model [%value%] ? ','model')
model = ask('\nAvailable models:\n\
  (1) Restoration   (Schori, 2018)\n\
  (2) Forward model (Schori, 2018)\n\
  (3) Restoration   (Affolter, 2003 & Philippe, 1995)\n\
  (4) Restoration   (Philippe, 1995)\n\n\
Which model do you want to use [%value%]?','model')

if model == 1:
    model_strike = 'models/strike_rest'
    model_length = 'models/length_rest'
    invert_strikes = False
elif model == 2:
    model_strike = 'models/strike_fwd'
    model_length = 'models/length_fwd'
    invert_strikes = True
elif model == 3:
    model_strike = 'models/Affolter_Philippe/Affolter_Philippe_strike_300_mc'
    model_length = 'models/Affolter_Philippe/Affolter_Philippe_length_300_mc'
    invert_strikes = False
elif model == 4:
    model_strike = 'models/Philippe/Philippe1995_strike1'
    model_length = 'models/Philippe/Philippe1995_length1'
    invert_strikes = False

# --- Mark compression / extension
print "\nWhich markings do you want to set?"
print "  (0) None"
print "  (1) Compression (pixel overlaps)"
print "  (2) Extension (empty pixels)"
print "  (3) Compression and extension"

markings = ask('\nYour choice [%value%] ? ','markings')
print

if markings == 0:
    mark_overlap = False
elif markings == 1:
    mark_overlap = True
elif markings == 2:
    mark_overlap = False
    bgcolor = (200,200,255)
elif markings == 3:
    mark_overlap = True
    bgcolor = (200,200,255)

# --- Loading all rasters
print "Map to unfold"
myImage = spatialImage(input_map)
print "Grid containing vector DIRECTION:"
myStrike = spatialImage(model_strike)
print "Grid containing vector LENGTH:"
myLength = spatialImage(model_length)

max_displacement = int(myLength.im.getextrema()[1])

# Looping through steps
# Make sure intervals are at rounded km steps
start_displacement = int(round(max_displacement / 1000,0)*1000)

# Sequence is backwards
if sequence == 1:

    # --- Calculate and draw unfolded version (=first step)
    sequence_number = 1
    myNewImage = deform_image(max_displacement, model_type, sequence_number)
    
    if stepsize != 0:
        for x in xrange(start_displacement, 0, -stepsize):
            sequence_number = sequence_number + 1
            myNewImage = deform_image(x, model_type, sequence_number)

# Forward sequence
if sequence == 0:

    # --- Calculate and draw unfolded version (=first step)
    sequence_number = 1
    myNewImage = deform_image(0, model_type, sequence_number)

    if stepsize != 0:
        for x in xrange(stepsize, start_displacement, stepsize):
            sequence_number = sequence_number + 1
            myNewImage = deform_image(x, model_type, sequence_number)

        
# --- Draw 0 displacement for completeness (having text imprint)
sequence_number = sequence_number + 1
myNewImage = deform_image(0, model_type, sequence_number = sequence_number)

# --- Create video
createVideo()

print "\nDone"


