# Spatial restoration / forward script
#   * Grids and shapefile must be in the same reference coordinate system
#   * 1. Shapefile with points to restore
#   * 2. Grid with strikes (direction of displacement)
#   * 3. Grid with amount of displacement (m) / vector size
#   * Grid 2 and 3 should have the same extent and resolution for best results
#
# -- Marc Schori, 2020
#
# Make sure that arcpy is installed in your version of python
# (Arcpy comes with ESRI ArcMap, normally installed on c:\python\ArcMapXX.X)
#___________________________________________________________

import numpy
import sys
from ConfigParser import SafeConfigParser
import glob
import ntpath
import os

# ----------------------------------------------
# Invert strikes
# ----------------------------------------------
def strike_inv(strike):
    strike = strike + 180
    if strike > 359:
        strike = strike - 360
    return strike

# --------------------------
# Function returning new position after adding vector
# using geographic circle with N = 0 degree
# -------------------------
def trans(x_coord, y_coord, strike, length):

    # Do not alter pixel if length or strike are not defined
    if length <= 0 or strike < 0:
        print "WARNING: < 0"
        return (x_coord,y_coord)

    # Invert strikes if specified
    if invert_strikes:
        strike = strike_inv(strike)
        
    alpha = strike * numpy.pi / 180
    x_new = x_coord + (numpy.sin(alpha) * length)
    y_new = y_coord + (numpy.cos(alpha) * length)
    
    return (x_new,y_new)

# -----------------------------------
# Store and read values from INI
# -----------------------------------
def read_value(key, section = 'main'):
    config = SafeConfigParser()
    config.read('config_vector.ini')

    try:
        s = config.get(section, key)
    except:
        s = ''
    
    return s

def store_value(key, value, section = 'main'):
    config = SafeConfigParser()
    config.read('config_vector.ini')

    try:
        config.add_section(section)
    except:
        pass

    try:
        config.set(section, key, str(value))
    except:
        print(sys.exc_info()[0])
        
    with open('config_vector.ini', 'w') as f:
        config.write(f)

# -----------------------------------
# Quick function for user interaction
# -----------------------------------
def ask(text, key):
    value = read_value(key)
    text = text.replace("%value%", value)

    x = raw_input(text)
    if x == '':
        x = value
    else:
        store_value(key,x)
        
    if x.isdigit():
        return int(x)

    return x
    
# ----------------------------------
# Move shapefile points to new location
# ----------------------------------
def shapefile_handler(source_shp):

    # Definition of target file
    target_shp =  "results/" + os.path.splitext(ntpath.basename(source_shp))[0] + "_restored.shp"
    target_vectors = "results/" + os.path.splitext(ntpath.basename(source_shp))[0] + "_vectors.shp"

    # Initialise arcpy libraries
    # should be moved into module level
    print "Import arcpy library ..."
    import arcpy
    from arcpy import env  
    from arcpy.sa import *
    arcpy.env.workspace = "C:/Temp"
    arcpy.env.overwriteOutput = True
    print "Check Spatial analyst ..."
    arcpy.CheckOutExtension("Spatial")

    # Make a copy of the shapefile
    arcpy.CopyFeatures_management(source_shp, target_shp)
    
    # Create a line featureclass for vectors
    arcpy.CreateFeatureclass_management(ntpath.dirname(target_vectors), ntpath.basename(target_vectors), "POLYLINE", "#", "DISABLED", "DISABLED", source_shp)
    cursor_line = arcpy.da.InsertCursor(target_vectors, ["SHAPE@"])
    spatial_reference = arcpy.Describe(source_shp).spatialReference

    # Browse through shapefile points and move according deformation model
    print ("Move vertices of LINE shapefile ..."),
    with arcpy.da.UpdateCursor(target_shp, ["OID@", "SHAPE@"]) as cursor:   
    #with arcpy.da.UpdateCursor(target_shp, ["SHAPE@"]) as cursor:   
    #for row in arcpy.da.UpdateCursor(target_shp, ["OID@", "SHAPE@"]):
        for row in cursor:
            print("Feature ID {}:".format(row[0]))
            partnum = 0
            # Step through each part of the line (Multiline)
            for part in row[1]:
                # Print the part number
                print("   Part {}:".format(partnum))
                 # Step through each vertex in the feature
                pntnum = 0

                for pnt in part:
                    if pnt:
                        # Print x,y coordinates of current point
                        #print("      {}, {}".format(pnt.X, pnt.Y))
                        
                        #------------------------------------------------------
                        # Start reading model and moving object here
                        #------------------------------------------------------
                        try:
                            # Read length and strike from model
                            length = float(arcpy.GetCellValue_management(model_length, str(pnt.X) + chr(32) + str(pnt.Y)).getOutput(0))
                            strike = float(arcpy.GetCellValue_management(model_strike, str(pnt.X) + chr(32) + str(pnt.Y)).getOutput(0))
                        except:
                            length = 0
                            strike  = 0
                            print("      WARNING: position in grid does not exist.")
                            if remove_pts == 1:
                                print("         -> Removing point")
                                cursor.deleteRow
                        else:
                            # Calculate new position                          
                            newpos = trans(pnt.X, pnt.Y, strike, length)
                            
                            # Add a vector to the POLYLINE layer
                            array = arcpy.Array([arcpy.Point(pnt.X, pnt.Y), arcpy.Point(newpos[0], newpos[1])])
                            polyline = arcpy.Polyline(array, spatial_reference)
                            cursor_line.insertRow([polyline])
                             
                            # Store new position to point layer
                            try:
                                ary = row[1].getPart(partnum)
                                ary.replace(pntnum,Point(newpos[0], newpos[1]))
                                newLine = arcpy.Polyline(ary)
                                row[1] = newLine
                            #row[1].setValue("SHAPE",newLine)
                            
                            #row[1].getPart(partnum)[pntnum].X = newpos[0]
                            #row[1].getPart(partnum)[pntnum].Y = newpos[1]
                                cursor.updateRow(row)
                            except:
                                print("      WARNING: Could not move vertex " + str(pntnum))
    
                    else:
                        # If pnt is None, this represents an interior ring
                        print("Interior Ring:")

                    pntnum += 1
                partnum += 1
                
                
                
                '''
                with arcpy.da.UpdateCursor(fc,['SHAPE@','NewX1','NewY1']) as cursor:
    for row in cursor:
        geofeat=row[0]
        geoarry=geofeat.getPart(0)
        oldX1=geoarry[0].X
        oldY1=geoarry[0].Y
        #print('{0},{1}'.format(oldX1,oldY1))
        geoarry[0].X=row[1]
        geoarry[0].Y=row[2]
        newply=arcpy.Polyline(geoarry)
        row[0]=newply
        cursor.updateRow(row)
                '''
            
# ========================================================
#                       MAIN
# ========================================================

print "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"
print "@         VECTOR RESTORATION        @"
print "@                and                @"
print "@         FORWARD MODELLING         @"
print "@               V1.0                @"
print "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
print "Hint: coordinates of shapefile must be in CH1903 / LV03\n"
os.chdir(os.path.dirname(__file__))

# 1. User input
#---------------
maps = glob.glob("shapefiles/*.shp")
i = 0
print "Available shapefiles:"
for map in maps:
    print "  (" + str(i) + ") " + ntpath.basename(map)
    i = i + 1

x = ask("\nWhich file do you want to use [%value%] ? ", 'map')
source_shp = maps[x]
source_shp = os.path.splitext(source_shp)[0]
print source_shp

# 2. Ask for forward or restoration modelling
#--------------------------------------------
model = ask('\nAvailable models:\n\
  (1) Restoration   (Schori, V1.00)\n\
  (2) Forward model (Schori, V1.00)\n\
  What model do you want to use [%value%]?','model')

if model == 1:
    model_strike = 'models/Schori_v100/strike_rest'
    model_length = 'models/Schori_v100/length_rest'
    invert_strikes = False
elif model == 2:
    model_strike = 'models/Schori_v100/strike_fwd'
    model_length = 'models/Schori_v100/length_fwd'
    invert_strikes = True


    
# Ask what to do with points in undefined grid positions
#--------------------------------------------
remove_pts = ask('\nWhat to do with point locations not defined in deformation grids?:\n\
  (1) Remove point\n\
  (2) Keep point and leave at location\n\n\
What is your decision [%value%]?','remove_pts')

# 3. Run restoration
#--------------------------------------------
print "\n_________________ Start ____________________\n"
shapefile_handler(source_shp + ".shp")
print "\nDone"